# hobo - portal to configure and deploy applications
# Copyright (C) 2015-2019  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

from django.urls import reverse_lazy
from django.views.generic import FormView

from hobo.environment.models import Authentic
from hobo.environment.utils import get_setting_variable
from .forms import SettingsForm, EnableForm


def get_variable(setting_name):
    return get_setting_variable(
        setting_name,
        service=Authentic.objects.get(secondary=False))

PLATFORMS = {
    'test': {
        'A2_FC_AUTHORIZE_URL': 'https://fcp.integ01.dev-franceconnect.fr/api/v1/authorize',
        'A2_FC_TOKEN_URL': 'https://fcp.integ01.dev-franceconnect.fr/api/v1/token',
        'A2_FC_USERINFO_URL': 'https://fcp.integ01.dev-franceconnect.fr/api/v1/userinfo',
        'A2_FC_LOGOUT_URL': 'https://fcp.integ01.dev-franceconnect.fr/api/v1/logout',
    },
    'prod': {
        'A2_FC_AUTHORIZE_URL': 'https://app.franceconnect.gouv.fr/api/v1/authorize',
        'A2_FC_TOKEN_URL': 'https://app.franceconnect.gouv.fr/api/v1/token',
        'A2_FC_USERINFO_URL': 'https://app.franceconnect.gouv.fr/api/v1/userinfo',
        'A2_FC_LOGOUT_URL': 'https://app.franceconnect.gouv.fr/api/v1/logout',
    }
}


class HomeView(FormView):
    template_name = 'hobo/franceconnect_home.html'
    form_class = SettingsForm
    success_url = reverse_lazy('franceconnect-home')

    def get_initial(self):
        initial = super(HomeView, self).get_initial()
        authorize_url = get_variable('A2_FC_AUTHORIZE_URL').value
        if authorize_url == 'https://fcp.integ01.dev-franceconnect.fr/api/v1/authorize':
            initial['platform'] = 'test'
        elif authorize_url == 'https://app.franceconnect.gouv.fr/api/v1/authorize':
            initial['platform'] = 'prod'

        initial['client_id'] = get_variable('A2_FC_CLIENT_ID').value
        initial['client_secret'] = get_variable('A2_FC_CLIENT_SECRET').value
        initial['scopes'] = get_variable('A2_FC_SCOPES').json or ['profile', 'email']

        return initial

    def form_valid(self, form):
        for key, value in PLATFORMS[form.cleaned_data['platform']].items():
            variable = get_variable(key)
            variable.value = value
            variable.save()

        variable = get_variable('A2_FC_CLIENT_ID')
        variable.value = form.cleaned_data['client_id']
        variable.save()

        variable = get_variable('A2_FC_CLIENT_SECRET')
        variable.value = form.cleaned_data['client_secret']
        variable.save()

        variable = get_variable('A2_FC_VERIFY_CERTIFICATE')
        variable.value = 'true'
        variable.save()

        variable = get_variable('A2_FC_USER_INFO_MAPPINGS')
        variable.value = json.dumps({
            'last_name': {
                'ref': 'family_name',
                'verified': True,
            },
            'first_name': {
                'ref': 'given_name',
                'verified': True,
            },
            'title': {
                'ref': 'gender',
                'translation': 'simple',
                'translation_simple': {
                    'male': 'Monsieur',
                    'female': 'Madame',
                },
                'verified': True,
            },
            'email': 'email',
        })
        variable.save()

        variable = get_variable('A2_FC_SCOPES')
        variable.json = form.cleaned_data['scopes']
        variable.save()

        return super(HomeView, self).form_valid(form)

    def get_context_data(self, **kwargs):
        context = super(HomeView, self).get_context_data(**kwargs)
        context['enabled'] = bool(get_variable('A2_FC_ENABLE').json)
        return context

home = HomeView.as_view()


class EnableView(FormView):
    form_class = EnableForm
    template_name = 'hobo/franceconnect_enable.html'
    success_url = reverse_lazy('franceconnect-home')

    def form_valid(self, form):
        variable = get_variable('A2_FC_ENABLE')
        variable.value = 'true'
        variable.save()
        return super(EnableView, self).form_valid(form)

enable = EnableView.as_view()


class DisableView(FormView):
    form_class = EnableForm
    template_name = 'hobo/franceconnect_disable.html'
    success_url = reverse_lazy('franceconnect-home')

    def form_valid(self, form):
        variable = get_variable('A2_FC_ENABLE')
        variable.value = 'false'
        variable.save()
        return super(DisableView, self).form_valid(form)

disable = DisableView.as_view()
