import os
import sys

from django.db import connection, transaction
from django.core.management.base import CommandError, BaseCommand

from hobo.multitenant.middleware import TenantMiddleware, get_tenant_model

class Command(BaseCommand):
    help = "Create tenant(s) by hostname(s)"

    def add_arguments(self, parser):
        parser.add_argument('hostnames', metavar='HOSTNAME', nargs='+')

    def handle(self, hostnames, **options):
        verbosity = int(options.get('verbosity'))
        if not hostnames:
            raise CommandError("you must give at least one tenant hostname")

        if '-' in hostnames: # get additional list of hostnames from stdin
            hostnames = list(hostnames)
            hostnames.remove('-')
            hostnames.extend([x.strip() for x in sys.stdin.readlines()])

        for hostname in hostnames:
            try:
                tenant_base = TenantMiddleware.base()
            except AttributeError:
                raise CommandError("you must configure TENANT_BASE in your settings")
            if not tenant_base:
                raise CommandError("you must set a value to TENANT_BASE in your settings")
            tenant_dir = os.path.join(tenant_base, hostname)
            if os.path.exists(tenant_dir):
                raise CommandError('tenant already exists')
            tenant_dir_tmp = os.path.join(
                    tenant_base, hostname + '__CREATION_IN_PROGRESS.invalid')
            try:
                os.mkdir(tenant_dir_tmp, 0o755)
            except OSError as e:
                raise CommandError('cannot start tenant creation (%s)' % str(e))
            # tenant creation in database
            try:
                connection.set_schema_to_public()
                schema = TenantMiddleware.hostname2schema(hostname)
                tenant = get_tenant_model()(schema_name=schema, domain_url=hostname)
                if verbosity >= 1:
                    print()
                    print(self.style.NOTICE("=== Creating schema ")
                        + self.style.SQL_TABLE(tenant.schema_name))
                with transaction.atomic():
                    tenant.create_schema(check_if_exists=True)
            except Exception as e:
                os.rmdir(tenant_dir_tmp)
                raise CommandError('tenant creation failed (%s)' % str(e))
            for folder in ('media', 'static', 'templates'):
                path = os.path.join(tenant_dir_tmp, folder)
                os.mkdir(path, 0o755)
            # activate the tenant by creating its directory
            os.rename(tenant_dir_tmp, tenant_dir)
